/*
 * FreeBSD firewire module for Python 2.3
 * Version 1.0
 * http://gdvrecv.sf.net/
 *
 * Copyright (c) 2004
 *      Henri Michelon
 * This product includes software developed by Hidetoshi Shimokawa.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: fwmodule.c,v 1.1 2004/11/01 20:47:10 hmichelon Exp $
 */
#include <fcntl.h>
#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/errno.h>
#include <sys/eui64.h>
#include <dev/firewire/firewire.h>
#include <dev/firewire/iec13213.h>
#include <arpa/inet.h>

#include <Python.h>

/* fwmodule_dv.c */
extern int dvrecv(int d, char ich, int count, PyObject *context,
                  PyObject *start_callback, PyObject *frame_callback, PyObject *eagain_callback);


/* module methods */
static PyObject *fw_open_dev(PyObject *self, PyObject *args);
static PyObject *fw_get_dev(PyObject *self, PyObject *args);
static PyObject *fw_get_crom(PyObject *self, PyObject *args);
static PyObject *fw_recv(PyObject *self, PyObject *args);

/* methods tables */
static PyMethodDef FirewireMethods[] = {
  {"open_dev",  fw_open_dev, METH_VARARGS, "Open a firewire device."},
  {"get_dev",  fw_get_dev, METH_VARARGS, "Get the list of firewire devices."},
  {"get_crom",  fw_get_crom, METH_VARARGS, "Return informations about a given node."},
  {"recv",  fw_recv, METH_VARARGS, "Receive a DV stream."},
  {NULL, NULL, 0, NULL}        /* Sentinel */
};

/* Default exception class */
static PyObject *FirewireError;


/* utility functions */

static PyObject* fw_trans_crom(u_int32_t *crom_buf);

static void fw_fweui2eui64(const struct fw_eui64 *fweui, struct eui64 *eui)
{
  *(u_int32_t*)&(eui->octet[0]) = htonl(fweui->hi);
  *(u_int32_t*)&(eui->octet[4]) = htonl(fweui->lo);
}


/* python module initialisation */
PyMODINIT_FUNC
initfirewire(void)
{
  PyObject *m = Py_InitModule("firewire", FirewireMethods);
  FirewireError = PyErr_NewException("firewire.error", NULL, NULL);
  Py_INCREF(FirewireError);
  PyModule_AddObject(m, "error", FirewireError);
}


static PyObject *
fw_open_dev(PyObject *self, PyObject *args)
{
  char *devbase;
  char devname[256];
  int i, fd;

  if (!PyArg_ParseTuple(args, "s", &devbase)) {
    return NULL;
  }

  for (i = 0; i < 4; i++) {
    snprintf(devname, sizeof(devname), "%s.%d", devbase, i);
    if ((fd = open(devname, O_RDWR)) >= 0) {
      break;
    }
  }
  if (fd < 0) {
    Py_INCREF(Py_None);
    return Py_None;
  }

  return Py_BuildValue("i", fd);
}


static PyObject *
fw_get_dev(PyObject *self, PyObject *args)
{
  int fd;
  struct fw_devlstreq *data;

  if (!PyArg_ParseTuple(args, "i", &fd)) {
    return NULL;
  }

  data = (struct fw_devlstreq *)malloc(sizeof(struct fw_devlstreq));
  if (data != NULL) {
    if (ioctl(fd, FW_GDEVLST, data) != -1) {
      int i;
      PyObject *result;
      if ((result = PyDict_New()) == NULL) {
        return NULL;
      }
      PyDict_SetItem(result, Py_BuildValue("s", "n"), Py_BuildValue("i", data->n));
      PyDict_SetItem(result, Py_BuildValue("s", "info_len"), Py_BuildValue("i", data->info_len));
      PyObject *devs = PyList_New(data->info_len);
      for (i = 0; i < data->info_len; i++) {
        PyObject *dev;
        PyObject *eui64;
        struct fw_devinfo *devinfo;

        devinfo = &data->dev[i];

        eui64 = PyDict_New();
        PyDict_SetItem(eui64, Py_BuildValue("s", "hi"), Py_BuildValue("i", devinfo->eui.hi));
        PyDict_SetItem(eui64, Py_BuildValue("s", "lo"), Py_BuildValue("i", devinfo->eui.lo));

        dev = PyDict_New();
        PyDict_SetItem(dev, Py_BuildValue("s", "eui"), eui64);
        PyDict_SetItem(dev, Py_BuildValue("s", "dst"), Py_BuildValue("i", devinfo->dst));
        PyDict_SetItem(dev, Py_BuildValue("s", "status"), Py_BuildValue("i", devinfo->status));

        PyList_SetItem(devs, i, dev);
      }
      PyDict_SetItem(result, Py_BuildValue("s", "dev"), devs);
      return result;
    }
  }
  Py_INCREF(Py_None);
  return Py_None;
}


static PyObject *fw_get_crom(PyObject *self, PyObject *args)
{
  int fd;
  PyObject *eui;
  struct fw_eui64 fweui;
  struct fw_crom_buf buf;
  u_int32_t crom_buf[1024/4];
  int len = 1024;

  if (!PyArg_ParseTuple(args, "iO", &fd, &eui)) {
    return NULL;
  }

  fweui.hi = PyLong_AsLong(PyDict_GetItemString(eui, "hi"));
  fweui.lo = PyLong_AsLong(PyDict_GetItemString(eui, "lo"));
  buf.eui = fweui;
  buf.len = len;
  buf.ptr = crom_buf;
  bzero(crom_buf, len);
  if (ioctl(fd, FW_GCROM, &buf) < 0) {
    return NULL;
  }

  return fw_trans_crom(crom_buf);
}


static void fw_parse_bus_info_block(PyObject *res, u_int32_t *p, int info_len)
{
  char addr[EUI64_SIZ];
  struct bus_info *bi;
  struct eui64 eui;

  bi = (struct bus_info *)p;
  fw_fweui2eui64(&bi->eui64, &eui);
  eui64_ntoa(&eui, addr, sizeof(addr));
  
  PyDict_SetItemString(res, "bus_name", Py_BuildValue("i", bi->bus_name));
  PyDict_SetItemString(res, "irmc", Py_BuildValue("i", bi->irmc));
  PyDict_SetItemString(res, "cmc", Py_BuildValue("i", bi->cmc));
  PyDict_SetItemString(res, "isc", Py_BuildValue("i", bi->isc));
  PyDict_SetItemString(res, "bmc", Py_BuildValue("i", bi->bmc));
  PyDict_SetItemString(res, "pmc", Py_BuildValue("i", bi->pmc));
  PyDict_SetItemString(res, "cyc_clk_acc", Py_BuildValue("i", bi->cyc_clk_acc));
  PyDict_SetItemString(res, "max_rec", Py_BuildValue("i", bi->max_rec));
  PyDict_SetItemString(res, "max_rom", Py_BuildValue("i", bi->max_rom));
  PyDict_SetItemString(res, "generation", Py_BuildValue("i", bi->generation));
  PyDict_SetItemString(res, "link_spd", Py_BuildValue("i", bi->link_spd));
  PyDict_SetItemString(res, "EUI64", Py_BuildValue("s", addr));
}


static PyObject* fw_trans_crom(u_int32_t *crom_buf)
{
  struct crom_context cc;
  static char *key_types = "ICLD";
  struct csrreg *reg;
  struct csrdirectory *dir;
  struct csrhdr *hdr;
  char info[256];
  u_int16_t crc;
  int len = sizeof(info);

  PyObject *res = PyDict_New();

  hdr = (struct csrhdr *)crom_buf;
  if (hdr->info_len == 1) {
    /* minimum ROM */
    struct csrreg *reg;
    reg = (struct csrreg *)hdr;
    PyDict_SetItemString(res, "vendorID", Py_BuildValue("i", reg->val));
    return res;
  }

  PyDict_SetItemString(res, "info_len", Py_BuildValue("i", hdr->info_len));
  PyDict_SetItemString(res, "crc_len", Py_BuildValue("i", hdr->crc_len));
  PyDict_SetItemString(res, "crc", Py_BuildValue("i", hdr->crc));
  crc = crom_crc(crom_buf+1, hdr->crc_len);
  if (crc == hdr->crc) {
    PyDict_SetItemString(res, "crc_ok", Py_BuildValue("s", "OK"));
  }
  fw_parse_bus_info_block(res, crom_buf+1, hdr->info_len);

  crom_init_context(&cc, crom_buf);
  dir = cc.stack[0].dir; 
  if (!dir) {
    /* no root directory - giving up */
    return res;
  }

  PyObject *pydir = PyDict_New();
  PyDict_SetItemString(pydir, "len", Py_BuildValue("i", dir->crc_len));
  PyDict_SetItemString(pydir, "crc", Py_BuildValue("i", dir->crc));
  crc = crom_crc((u_int32_t *)&dir->entry[0], dir->crc_len);
  if (crc == dir->crc) {
    PyDict_SetItemString(pydir, "crc_ok", Py_BuildValue("s", "OK"));
  }
  PyDict_SetItemString(res, "root_directory", pydir);
  if (dir->crc_len < 1) {
    return res;
  }

  PyObject *regs = PyList_New(0);
  while (cc.depth >= 0) {
    PyObject *pyreg;
    reg = crom_get(&cc);
    pyreg = PyDict_New();
    PyDict_SetItemString(pyreg, "depth", Py_BuildValue("i", cc.depth));
    PyDict_SetItemString(pyreg, "key", Py_BuildValue("i", reg->key));
    PyDict_SetItemString(pyreg, "key_type_c", Py_BuildValue("c", key_types[(reg->key & CSRTYPE_MASK)>>6]));
    PyDict_SetItemString(pyreg, "key_type_i", Py_BuildValue("i", reg->key & CSRKEY_MASK));
    PyDict_SetItemString(pyreg, "val", Py_BuildValue("i", reg->val));
    switch (reg->key) {
    /*case 0x13:
              crom_desc_specver(0, reg->val, info, len);
              PyDict_SetItemString(pyreg, "info", Py_BuildValue("s", info));
              break;*/
    case 0x81:
              crom_parse_text(&cc, info, len);
              PyDict_SetItemString(pyreg, "info", Py_BuildValue("s", info));
              break;
    }
    PyList_Append(regs, pyreg);
    crom_next(&cc);
  }
  PyDict_SetItemString(res, "regs", regs);
  return res;
}


static PyObject *fw_recv(PyObject *self, PyObject *args)
{
  int fd;
  PyObject *context;
  PyObject *py_start_callback;
  PyObject *py_frame_callback;
  PyObject *py_eagain_callback;
  PyObject *result = NULL;

  if (!PyArg_ParseTuple(args, "iOOOO", &fd,
                        &context,
                        &py_start_callback,
                        &py_frame_callback,
                        &py_eagain_callback)) {
    return NULL;
  }
  
  if ((!PyCallable_Check(py_start_callback)) ||
      (!PyCallable_Check(py_frame_callback)) ||
      (!PyCallable_Check(py_eagain_callback))) {
    PyErr_SetString(PyExc_TypeError, "parameter must be callable");
    return NULL;
  }
  /* Boilerplate to return "None" */
  Py_INCREF(Py_None);
  result = Py_None;

#define TAG     (1<<6)
#define CHANNEL 63
  dvrecv(fd, TAG | CHANNEL, -1, context, py_start_callback, py_frame_callback, py_eagain_callback);

  return result;
}
