/*
 * Copyright (C) 2003
 * 	Hidetoshi Shimokawa. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 *	This product includes software developed by Hidetoshi Shimokawa.
 *
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $FreeBSD: src/usr.sbin/fwcontrol/fwdv.c,v 1.5 2003/04/17 03:38:03 simokawa Exp $
 * $Id: fwmodule_dv.c,v 1.1 2004/11/01 20:47:10 hmichelon Exp $
 */
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/uio.h>

#if __FreeBSD_version >= 500000
#include <arpa/inet.h>
#endif

#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include <dev/firewire/firewire.h>
#include <dev/firewire/iec68113.h>

#include <Python.h>

#define DEBUG		0
#define FIX_FRAME	1

struct frac {
	int n,d;	
};

struct frac frame_cycle[2]  = {
	{8000*100, 2997},	/* NTSC 8000 cycle / 29.97 Hz */
	{320, 1},		/* PAL  8000 cycle / 25 Hz */
};
int npackets[] = {
	250		/* NTSC */,
	300		/* PAL */
};
struct frac pad_rate[2]  = {
	{203, 2997},	/* = (8000 - 29.97 * 250)/(29.97 * 250) */
	{1, 15},	/* = (8000 - 25 * 300)/(25 * 300) */
};
char *system_name[] = {"NTSC", "PAL"};
int frame_rate[] = {30, 25};

#define PSIZE 512
#define DSIZE 480
#define NCHUNK 8

#define NPACKET_R 256
#define NPACKET_T 255
#define TNBUF 100	/* XXX too large value causes block noise */
#define NEMPTY 10	/* depends on TNBUF */
#define RBUFSIZE (PSIZE * NPACKET_R)
#define MAXBLOCKS (300)
#define CYCLE_FRAC 0xc00

void my_writev(const struct iovec *wbuf, int vec,
               PyObject *context, PyObject *frame_callback)
{
  int i;
  PyObject *arglist;
  PyObject *buffer_list;
  buffer_list = PyList_New(vec);
  for (i = 0 ; i < vec; i++) {
    PyObject *buf = PyBuffer_FromMemory(wbuf[i].iov_base, wbuf[i].iov_len);
    PyList_SET_ITEM(buffer_list, i, buf);
  }
  arglist = Py_BuildValue("(OO)", context, buffer_list);
  PyEval_CallObject(frame_callback, arglist);
  Py_DECREF(arglist);
  Py_DECREF(buffer_list);
}

int dvrecv(int d, char ich, int count,
           PyObject *context, PyObject *start_callback,
           PyObject *frame_callback, PyObject *eagain_callback)
{
	struct fw_isochreq isoreq;
	struct fw_isobufreq bufreq;
	struct dvdbc *dv;
	struct ciphdr *ciph;
	struct fw_pkt *pkt;
	char *pad, *buf;
	u_int32_t *ptr;
	int len, tlen, npad, k, m, vec, system = -1, nb;
	int nblocks[] = {250 /* NTSC */, 300 /* PAL */};
	struct iovec wbuf[NPACKET_R];
        PyObject *arglist;
        int started = 0;

	bufreq.rx.nchunk = NCHUNK;
	bufreq.rx.npacket = NPACKET_R;
	bufreq.rx.psize = PSIZE;
	bufreq.tx.nchunk = 0;
	bufreq.tx.npacket = 0;
	bufreq.tx.psize = 0;
	if (ioctl(d, FW_SSTBUF, &bufreq) < 0) {
          return 0;
	}

	isoreq.ch = ich & 0x3f;
	isoreq.tag = (ich >> 6) & 3;

	if( ioctl(d, FW_SRSTREAM, &isoreq) < 0) {
          return 0;
        }

	buf = (char *)PyMem_Malloc(RBUFSIZE);
	pad = (char *)PyMem_Malloc(DSIZE*MAXBLOCKS);
	memset(pad, 0xff, DSIZE*MAXBLOCKS);
	bzero(wbuf, sizeof(wbuf));

	k = m = 0;
	while (count <= 0 || k <= count) {
		tlen = len = read(d, buf, RBUFSIZE);
		if (len < 0) {
			if (errno == EAGAIN) {
                                if (started) {
                                  PyMem_Free(buf);
                                  PyMem_Free(pad);
                                  return 0;
                                }
                                arglist = Py_BuildValue("(O)", context);
                                PyEval_CallObject(eagain_callback, arglist);
                                Py_DECREF(arglist);
				if (len <= 0)
					continue;
			} else {
                                PyMem_Free(buf);
                                PyMem_Free(pad);
                                return 0;
                        }
		}
		vec = 0;
		ptr = (u_int32_t *) buf;
again:
		pkt = (struct fw_pkt *) ptr;	

		ciph = (struct ciphdr *)(ptr + 1);	/* skip iso header */
		if (ciph->fmt != CIP_FMT_DVCR) {
                                  PyMem_Free(buf);
                  return 0;
                }
		ptr = (u_int32_t *) (ciph + 1);		/* skip cip header */

		if (pkt->mode.stream.len <= sizeof(struct ciphdr))
			/* no payload */
			goto next;
		for (dv = (struct dvdbc *)ptr;
				(char *)dv < (char *)(ptr + ciph->len);
				dv+=6) {

			if  (dv->sct == DV_SCT_HEADER && dv->dseq == 0) {
				if (system < 0) {
                                  started = 1;
				  system = ciph->fdf.dv.fs;
                                  arglist = Py_BuildValue("(Os)", context, system_name[system]);
                                  PyEval_CallObject(start_callback, arglist);
                                  Py_DECREF(arglist);
				}

				/* Fix DSF bit */
				if (system == 1 &&
					(dv->payload[0] & DV_DSF_12) == 0)
					dv->payload[0] |= DV_DSF_12;
				nb = nblocks[system];
#if FIX_FRAME
				if (m > 0 && m != nb) {
					/* padding bad frame */
					npad = ((nb - m) % nb);
					if (npad < 0)
						npad += nb;
					npad *= DSIZE;
					wbuf[vec].iov_base = pad;
					wbuf[vec++].iov_len = npad;
					if (vec >= NPACKET_R) {
						my_writev(wbuf, vec, context, frame_callback);
						vec = 0;
					}
				}
#endif
				k++;
				m = 0;
			}
			if (k == 0 || (count > 0 && k > count))
				continue;
			m++;
			wbuf[vec].iov_base = (char *) dv;
			wbuf[vec++].iov_len = DSIZE;
			if (vec >= NPACKET_R) {
				my_writev(wbuf, vec, context, frame_callback);
				vec = 0;
			}
		}
		ptr = (u_int32_t *)dv;
next:
		if ((char *)ptr < buf + tlen)
			goto again;
		if (vec > 0) {
			my_writev(wbuf, vec, context, frame_callback);
                }
	}
        PyMem_Free(buf);
        PyMem_Free(pad);
	return 0;
}


